import numpy as np
import pygoko

tree = pygoko.CoverTree()
tree.load_yaml_config("../../data/ember_complex.yml")
tree.fit()


print(tree.knn(tree.data_point(0),5))

print("============= TRACE =============")
trace = tree.path(tree.data_point(59999))
for distance,address in trace:
    node = tree.node(address)
    mean = node.cover_mean()
    if mean is not None:
        print(f"\tNode {node.address()}, mean: {mean.mean()}")
    else:
        print(f"\tNode {node.address()}, MEAN IS BROKEN")

print("============= KL Divergence =============")
prior_weight = 1.0
observation_weight = 1.3
sequence_cap = 10
sequence_len = 20
sequence_count = 10
normal_stats = tree.kl_div_dirichlet_basestats(prior_weight,
    observation_weight,
    sequence_len,
    sequence_count,
    sequence_cap)
for i,stats in enumerate(normal_stats[0]):
    print(stats)
print("============= KL Divergence Normal Use =============")
kl_tracker = tree.kl_div_dirichlet(prior_weight,observation_weight,sequence_cap)
for i in range(200):
    kl_tracker.push(tree.data_point(i))
    print(kl_tracker.stats())


print("============= KL Divergence Attack =============")

kl_attack_tracker = tree.kl_div_dirichlet(prior_weight,observation_weight,sequence_cap)
for i in range(10):
    kl_attack_tracker.push(tree.data_point(i))
    print(kl_attack_tracker.stats())
for i in range(10):
    kl_attack_tracker.push(tree.data_point(0))
    print(kl_attack_tracker.stats())




import pygoko

import numpy as np
from one_d_viz import show1D


data = np.array([[0.499], [0.48], [-0.49], [0.0]],dtype=np.float32)

tree = pygoko.CoverTree()
tree.set_scale_base(2)
tree.set_leaf_cutoff(0)
tree.fit(data)
show1D(tree,data)


import numpy as np
from sklearn.neighbors import KDTree
import pygoko
import pandas as pd

data = np.memmap("../../data/mnist.dat", dtype=np.float32)
data = data.reshape([-1,784])

tree = pygoko.CoverTree()
tree.load_yaml_config("../../data/mnist_complex.yml")
tree.set_leaf_cutoff(0)
tree.set_scale_base(1.2)
tree.set_min_res_index(-30)
tree.fit()

print(tree.knn(data[0],5))

for i,layer in enumerate(tree.layers()):
    print(f"On layer {layer.scale_index()}")
    if i < 2:
        for node in layer.nodes():
            print(f"\tNode {node.address()} mean: {node.cover_mean().mean()}")

print("============= TRACE =============")
trace = tree.path(data[59999])
for address in trace:
    print(address)
    node = tree.node(address[1])
    mean = node.cover_mean()
    if mean is not None:
        print(f"\tNode {node.address()}, mean: {mean.mean()}, summary: {node.label_summary()}")
    else:
        print(f"\tNode {node.address()}, MEAN IS BROKEN")

print("============= KL Divergence =============")
normal_stats = tree.kl_div_dirichlet_basestats(1.0,1.3,100,10,20)
for i,vstats in enumerate(normal_stats[:1]):
    for stats in vstats:
        print(stats)
print("============= KL Divergence Normal Use =============")
kl_tracker = tree.kl_div_dirichlet(1.0,1.3,20)
for x in data[:50]:
    kl_tracker.push(x)
    print(kl_tracker.stats())


print("============= KL Divergence Attack =============")

kl_attack_tracker = tree.kl_div_dirichlet(1.0,1.3,20)
for i in range(50):
    kl_attack_tracker.push(data[0])
    print(kl_attack_tracker.stats())



import numpy as np
from sklearn.neighbors import KDTree

tree = KDTree(data, leaf_size=2)
dist, ind = tree.query(data[:10], k=5)
print(dist[0])
print(ind[0])

dist, ind = tree.query(np.zeros([1,784],dtype=np.float32), k=5)
print(dist[0])
print(ind[0])

nbrs = {"d0":dist[:,0],
        "d1":dist[:,1],
        "d2":dist[:,2],
        "d3":dist[:,3],
        "d4":dist[:,4],
        "i0":ind[:,0],
        "i1":ind[:,1],
        "i2":ind[:,2],
        "i3":ind[:,3],
        "i4":ind[:,4],}

csv = pd.DataFrame(nbrs)
csv.to_csv("../../data/mnist_nbrs.csv")


import pygoko

import numpy as np
from math import pi
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.path as mpath
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
from matplotlib.collections import PatchCollection
from matplotlib import cm
cmap= plt.get_cmap("jet")
norm = mpl.colors.Normalize(vmin=0.0, vmax=1.0)

def show1D(tree,data):
    patches = []
    lines = []
    x_center = []
    y_center = []
    top = tree.top_scale()
    bottom = tree.bottom_scale()
    for i in range(top,bottom,-1):
        layer = tree.layer(i)
        width = layer.radius()
        scale_index = layer.scale_index()

        y = 1.5 + 0.1*scale_index
        point_indexes, center_points = layer.centers()
        for my_pi, c in zip(point_indexes,center_points):
            x = c[0]
            x_center.append(x)
            y_center.append(y)
            child_addresses = layer.child_addresses(my_pi)
            if layer.is_leaf(my_pi):
                patches.append(mpatches.Rectangle([x - width, y-0.05],2*width, 0.08, ec="none",color="orange"))

                lines.append(mlines.Line2D([x,x], [y,0.0],color="blue"))

            else:
                patches.append(mpatches.Rectangle([x - width, y-0.05],2*width, 0.08, ec="none",color="blue"))
                for si,pi in child_addresses:
                    y_next = 1.5+0.1*si
                    x_next = data[pi][0]
                    lines.append(mlines.Line2D([x,x_next], [y,y-0.05],color="blue"))
                    lines.append(mlines.Line2D([x_next,x_next], [y-0.05,y_next],color="blue"))

            for c in layer.singleton_points(my_pi):
                lines.append(mlines.Line2D([c[0],c[0]], [y-0.05,0.0],color="orange"))
                lines.append(mlines.Line2D([x,c[0]], [y,y-0.05],color="orange"))

    fig, ax = plt.subplots()
    ax.set_xlim((-1.01,1.01))
    ax.set_ylim((-0.1,2.1))
    ax.scatter(data[:,0].reshape((data.shape[0],)),np.zeros([data.shape[0]]),color="orange")
    collection = PatchCollection(patches,match_original=True, alpha=0.3)
    for l in lines:
        ax.add_line(l)
    ax.scatter(x_center,y_center)
    ax.add_collection(collection)
    plt.show()
    
if __name__ == '__main__':
    numPoints = 42

    data1 = np.random.normal(loc=0.5,scale=0.2,size=(numPoints//3,)).reshape(numPoints//3,1)
    data2 = np.random.normal(loc=-0.5,scale=0.1,size=(2*numPoints//3,)).reshape(2*numPoints//3,1)
    data = np.concatenate([data1,data2],axis=0)
    data = np.concatenate([data,np.zeros([numPoints,1])],axis=1).astype(np.float32)

    tree = pygoko.CoverTree()
    tree.set_scale_base(2)
    tree.set_leaf_cutoff(0)
    tree.fit(data)
    show1D(tree,data)